#!/usr/bin/env tclsh
#####################################################################
;# NAME:   pdf.tcl
;# AUTHOR: Clif Flynt
;# DATE:   3/2004
;# DESC:   A trivial PDF generator library.
;#   This code does not use 
;#   compression, Filters, transparency, or other fancy stuff.
;# 
;#   It writes data to a file, and uses [tell] to track the number of
;#   bytes written.  Conversion to something that will return a
;#   (possibly huge) string is left as an exercise for the user.
;#
;# PARAMETERS:  
;#
;#
#####################################################################
# This code is written and copyrighted by Clif Flynt, 2002
#
#   Placed in the public domain by:
#   Noumena Corporation
#   9300 Fleming Rd.
#   Dexter, MI  48130
#   734-426-1066
#   http://www.noucorp.com
#   Contact: clif@noucorp.com
#
# It is licenced for public use with these terms:
#
#  This licence agreement must remain intact with the code.
# 
#  You may use this code freely for commercial and non-commercial purposes.
#
#  You may examine and modify the code to your heart's content, but are
#  requested to make your work to available to the public.
#
#  You may not claim to have authored the code (aside from modifications
#  you may have made.)
#
#  You may not distribute the code without this license agreement
#
#####################################################################
#
# Procedure naming convention for this package:
#   proc names that start with UpperCase letters are private.
#   procs that modify the PDF file end in _Write
#   procs that return a string to append to a stream object
#     are a simple name like "pdfRectangle", "pdfText", or "pdfLine".
#   
#   Introduction to PDF basics
# Per PDF Reference guide 1.5 V5 P49
# 
# A PDF document is a set of objects in a flattened format.
# 
#   The important one is the stream, that's what describes all the graphic
# objects being displayed.
#   
#   File Format is Described in the PDF Referrence Manual, Section 3.4 :
# This code's specific format is:
#   %PDF-1.x
#   Definition objects: Catalog, Outline Pages Page, Fonts, etc.
#   Page objects  Things to be drawn, text, graphics, etc.
#   xref table - List of offsets of previous objects
#   Trailer
#   
# An object can be
#   Null		Primitive		null
# 
#   Boolean	Primitive		true false
# 
#   Numeric
#     Integer     Primitive		1 +4 -4 0
#     Real        Primitive		1.1 +4.2 -3.14 0.0
# 
#   String
#     Literal	Enclosed in paren	(A string may 
#                                          contain newlines)
#         Strings may also contain escaped special characters:
#         \n    Line feed (LF)
#         \r    Carriage return (CR)
#         \t    Horizontal tab (HT)
#         \b    Backspace (BS)
#         \f    Form feed (FF)
#         \(    Left parenthesis
#         \)    Right parenthesis
#         \\    Backslash
#         \ddd    Character code ddd (octal)
# 
#     Hex-encoded Enclosed in angles      <6520> -- "A "
# 
#   Name          References some object
#                 Starts with /            /AName, /A#20Space
# 
#   Array         Heterogynous list
#                 Enclosed in []           [123 3.14 false (string) /name]
# 
#   Dictionary    Key/Value pairs, not a named data structure, 
#                 but a part of some other complex object.
# 		/Type and /SubType keys must have Name for value.
#                 Enclosed in << >>
# 		                         <</Pi 3.14
# 					   /Type /Example
# 					 >>
#   Stream	Complex
# 
#                 per P 58
# A stream object, like a string object, is a sequence of bytes. However,
# a PDF appli- cation can read a stream incrementally, while a string must
# be read in its entirety. Furthermore, a stream can be of unlimited
# length, whereas a string is subject to an implementation limit. For this
# reason, objects with potentially large amounts of data, such as images
# and page descriptions, are represented as streams. 
# 
# Note: As with strings, this section describes only the syntax for
# writing a stream as a sequence of bytes. What those bytes represent is
# determined by the context in which the stream is referenced. 
# 
# A stream consists of a dictionary that describes a sequence of bytes,
# followed by zero or more bytes bracketed between the keywords stream and
# endstream: 
# 
# dictionary
# stream
# ...Zero or more bytes...
# endstream
#       A string dictionary must include a Length entry.
#         Length		How many bytes will be seen before the endstream
# 
# 
# An indirect object is basicly a "named" object (named with an integer)

# The definition of an indirect object in a PDF file consists of its
# object number and generation number, followed by the value of the object
# itself bracketed between the keywords obj and endobj. For example, the
# definition 
# 
# 12 0 obj
# (Brillig)
# endobj
# 
# defines an indirect string object with an object number of 12, a
# generation number of 0, and the value Brillig. 
# 
# The PDF file generated by this script consists of several
# indirect objects.
#
# The initial definition descriptors, font descriptors, etc are
# each arrays in an indirect object.  
#
# The drawing elements are a stream, in the last indirect object
# 
# The final items in the file is the xref table (which is not an object)
# the trailer info, and the pointer to the start of the xref table.
#
# The xref table contains the byte offsets of the indirect objects.
# It's a sort of Table of Contents, or Directory.
#
#
# The flow for using this library is:
# Initialize the PDF output (initPDF_Write fileName)
# Generate graphic streams, possibly adding Font Objects as fonts are
#   identified.
# Write the contents tree with: pdf::docTree_Write
# Write the start directive: pdf::directive_Write /PDF /Text
# Write the graphics stream object: streamObj_Write
# Close the file with finish_Write

package provide pdf 2.0

namespace eval pdf {
set PDF(fmtStr) {%010d %05d %s }

    variable var

    # Multiplication factor for splines in ovals.
    set var(kappa) [expr {(4.0 * (sqrt(2.0) - 1.0)) / 3.0}]
    # Various factors of pi
    set var(piDiv2) [expr {acos(0.0)}]
    set var(pi) [expr {2.0 * $var(piDiv2)}]
    set var(2pi) [expr {2.0 * $var(pi)}]
    set var(3piDiv2) [expr {$var(2pi) - $var(piDiv2)}]
    # Conversion multiplier from degrees to radians.
    set var(deg2rad) [expr {$var(pi) / 180.0}]

################################################################
# proc Color2PDF {color}--
#    Convert a color name or #rgb value to a percent r g b
#    postscript format
# Arguments
#   color	The name of the color, either a known color
#               name like "red", or a #rgb value
# 
# Results
#   No side effects

proc Color2PDF {color} {
    # If running in Tk, use winfo to parse color
    # else use a somewhat kludgier approach 
  if {[namespace exists ::tk]} {
    foreach {r g b} [split [winfo rgb . $color]] {}
    set max 65536.0
  } else {
    # Not running in a tk application, so we punt
    #  on color values.  A few hardcoded common names,
    #  or parse the #rgb value, or throw an error.
    array set lkup {
	yellow  {1.0 1.0 0.0} 
	cyan    {0.0 1.0 1.0}
	magenta {1.0 0.0 1.0}
    red     {1.0 0.0 0.0}
	green   {0.0 1.0 0.0} 
	blue    {0.0 0.0 1.0} 
	orange  {1.0 0.6 0.0}
	purple  {0.6 0.2 0.9}
	violet  {0.9 0.5 0.9}
	white   {1.0 1.0 1.0} 
    black   {0.0 0.0 0.0} 
	gray    {0.8 0.8 0.8} 
	gray75  {0.7 0.7 0.7} 
	gray50  {0.5 0.5 0.5} 
	gray25  {0.2 0.2 0.2} 
	pink    {1.0 0.7 0.8}
	}
    
    # If a common (known!) color name, just return.
    # There's nothing more interesting to be done.

    if {[info exists lkup($color)]} {
        return $lkup($color)
    }

    if {[string first # $color] == 0} {
      set colorlen [expr {[string length $color] / 3}]
      
      set start 1
      set end $colorlen
      
      foreach cc {r g b} {
        set $cc "0x[string range $color $start $end]"
        incr start $colorlen
        incr end $colorlen
      }
    } else {
      error "Unknown Color" "$color is not a known color value"
    }
    
    set max [expr {pow(16,$colorlen)-1}]

  }

  foreach c {r g b} {
    append clr [format "%4.2f" [expr [set $c] / $max]]
    append clr " "
  }
  return $clr
}

proc ObjStart {} {
    variable PDF

    set pos [llength $PDF(xrefs)]
    lappend PDF(xrefs) [format $PDF(fmtStr) [tell $PDF(of)] $PDF(idN) $PDF(id)]
    
    puts $PDF(of) "$pos 0 obj"
    return $pos
}

proc ObjEnd {} {
    variable PDF
    puts $PDF(of) "endobj"
}

################################################################
# proc pdfText {font size x y txt}--
#    Returns a text stream object at an X/Y location in given font
#    as an string, which can be included in a Stream Object.
#
#    Modifies the original $txt data by escaping special characters
#
# PDFReferece Guide 1.5 V5 P 48:
#
# The delimiter characters (, ), <, >, [, ], {, }, /, and % are special.
# They delimit syntactic entities such as strings, arrays, names, and
# comments. Any of these characters terminates the entity preceding it and
# is not included in the entity. 

# Empirical evidence suggests ' should also be escaped.

# Arguments
#   font  The font, in PDF format, (must reference a Font object)
#   size  Point size of text
#   x     X location in PDF document coords
#   y     Y location in PDF document coords
#   txt   Plain text
# 
# Results
#   Returns a string with appropriate prefix, postfix and 
#   escapes
# 
# Test msg:
#
# This is a {curly} and (paren) and <angle> and even
# [bracketed] piece of trash/garbage, don't you know,
# It's got a small % chance of succeeding.

proc pdfText {font size color x y txt} {
        set lst { \{ \} ( ) [ ] < > / % '}
        append str "BT\n"
        append str [color $color -nonstroke]
        append str "$font $size Tf\n"
        append str "$x $y Td\n"

        foreach l $lst {
	    regsub -all "\\$l" $txt "\\\\$l" txt
	}

    append str "($txt) Tj\n"
	append str "ET\n"
	return $str
}

proc pdfTransformatedText {font size color x y txt transformString} {
        set lst { \{ \} ( ) [ ] < > / % }
        append str "BT\n"
        append str "$transformString\n"
        append str [color $color -nonstroke]
        append str "$font $size Tf\n"
        append str "$x $y Td\n"

        foreach l $lst {
	    regsub -all "\\$l" $txt "\\\\$l" txt
	}

    append str "($txt) Tj\n"
	append str "ET\n"
	return $str
}

################################################################
# proc pdfText {font array size x y txt}--
#    Returns a text stream object at an X/Y location in given font
#    as an string, which can be included in a Stream Object.
#
#    Modifies the original $txt data by escaping special characters
#  and reversing the order, displaying the text from a right margin
#  to the left.
#
# PDFReferece Guide 1.5 V5 P 48:
#
# The delimiter characters (, ), <, >, [, ], {, }, /, and % are special.
# They delimit syntactic entities such as strings, arrays, names, and
# comments. Any of these characters terminates the entity preceding it and
# is not included in the entity. 

# Empirical evidence suggests ' should also be escaped.

# Arguments
#   font  The font, in PDF format, (must reference a Font object)
#   arrayName An array with character size lookuptable.
#   size  Point size of text
#   x     X location in PDF document coords
#   y     Y location in PDF document coords
#   txt   Plain text
# 
# Results
#   Returns a string with appropriate prefix, postfix and 
#   escapes
# 
# Test msg:
#
# This is a {curly} and (paren) and <angle> and even
# [bracketed] piece of trash/garbage, don't you know,
# It's got a small % chance of succeeding.

proc pdfTextRightJustify {font arrayName size color x y txt} {
        upvar $arrayName lookup

        set lst { \{ \} ( ) [ ] < > / % '}
	append str "BT\n"
        append str [color $color -nonstroke]
        append str "$font $size Tf\n"
        append str "$x $y Td\n"
	
	set cmd "$lookup(font.$font.[string range $txt end end]) "
	
	set first -40

        while {![string equal "" $txt]} {   
          set ch [string range $txt end end]
          set nxch [string range $txt end-1 end-1] 
          set ch1sz $lookup(font.$font.$ch)
	  if {![string equal "" $nxch]} {
              set ch2sz $lookup(font.$font.$nxch)
	  } else {
              set ch2sz 0
	  }
          set offset [expr {$ch1sz + $ch2sz + $first}]
          if {[lsearch $lst $ch] >= 0} {
	    set ch "\\$ch"
	  }
          append cmd "($ch) $offset "
          set txt [string range $txt 0 end-1]
	  set first 0
        }
        set cmd [format {[%s] TJ} $cmd]

        append str "$cmd\n"
	append str "ET\n"
	return $str
}

################################################################
# proc color {color args}--
#    Returns a string that assigns the Stroke color (default)
#    or non stroking color.
# Arguments
#   color - Color in Tcl terms (name or #rrggbb
#   ?-nonstroke? - Return non-stroke command instead of stroke
# Results
#   No side effects

proc color {color args} {
   set pdfColor [Color2PDF $color]
   if {[string equal $args -nonstroke]} {
     set cmd rg
   } else {
     set cmd RG
   }
   return "$pdfColor $cmd\n"
}


################################################################
# proc List2Dict {lst {depth 0} }--
#    Converts a Tcl list into a PDF dict
#    Adds indentation eye-candy that is not required by PDF spec
# Arguments
#   lst - A Tcl list of values.  Any sublist starting with / is treated 
#         as a sub-dictionary
#   depth - The depth of the current dictionary - for the indent
# Results
#   Returns a string that can be included in an object, etc.
#   No side effects

proc List2Dict {lst {depth 0} } {

   if {([llength $lst] & 1)} {
       #puts "LIST: $lst"
       error "bad list, must be List2Dict key1 val1 key2 val2 ..."
   }

   set prefix1 [string repeat { } $depth]
   incr depth 2
   set prefix2 [string repeat { } $depth]

   set dict "$prefix1<<"

   foreach {k v} $lst {
     if {([llength $v] > 1) && ([string first "/" [lindex $v 0]] == 0)} {
       # Treat /Font and /XObject different - they aren't nested 

       switch -glob [lindex $v 0] {
         /F[0-9]		-
	 /Im[0-9] 	{
           append dict "\n$prefix2$k  [list2Tbl $v $depth]"
	 }
         default {
           append dict "\n$prefix2$k  [List2Dict $v $depth]"
	 }
       }
     } else {
      append dict [format {%s%s%-20s%s} "\n" $prefix2 $k $v]
     }
   }
   append dict "\n$prefix1>>"
   return $dict
}

proc list2Tbl {lst {depth 0} } {

   if {([llength $lst] & 1)} {
       #puts "LIST: $lst"
       error "bad list, must be list2Tbl key1 val1 val2 val3 ..."
   }

   set prefix1 [string repeat { } $depth]
   incr depth 2
   set prefix2 [string repeat { } $depth]

   set dict "$prefix1<<"

   foreach {k v1 v2 v3} $lst {
      append dict [format {%s %s %s %s %s %s} "\n" $prefix2 $k $v1 $v2 $v3]
   }
   append dict "\n$prefix1>>"
   return $dict
}

################################################################
# proc pageObj_Write {geom }--
#    Writes a Stream Object (graphics and text drawing commands)
#    to the disk.
# Arguments
#   geom - A set of text and geometric glyph drawing commands
#
# Results
#   Updates the xref table,
#   Writes to disk.

proc pageObj_Write {geom args} {
    variable PDF
    
    set str $geom\n
    
    set lst $args
    lappend lst /Length [string length $str]
    
    set objNum [ObjStart]
    puts $PDF(of) [List2Dict $lst]
    puts $PDF(of) "stream"
    puts $PDF(of) $str
    puts $PDF(of) "endstream"
    ObjEnd
    return $objNum
}

################################################################
# proc streamObj_Write {stream args}--
#    Writes a Stream Object 
#    to the disk.
# Arguments
#   stream	: A stream to be written
#   args	: key/values pairs to proceed the stream data
# Results
#   Updates the xref table,
#   Writes to disk.

proc streamObj_Write {stream args} {
    variable PDF
    
    set str $stream\n
    set lst $args
    lappend lst /Length [string length $str]
    
    set objNum [ObjStart]
    puts $PDF(of) [List2Dict $lst]
    puts $PDF(of) "stream"
    puts $PDF(of) $str
    puts $PDF(of) "endstream"
    ObjEnd
    return $objNum
}
   
################################################################
# proc getFontName {local}--
#    Returns a PDF name for a font in 'local' naming convention
# Arguments
#   A local format name - /Times, /Arial, etc.
# 
# Results
#   Returns a string like /F1, /F2, etc.
# 
proc getFontName {local} {
    variable  PDF
    set pos [lsearch $PDF(fontList) $local]
    if {$pos < 0} {return ""}
    incr pos
    return [lindex $PDF(fontList) $pos]
}

################################################################
# proc getImageName {local}--
#    Returns a PDF name for a font in 'local' naming convention
# Arguments
#   A local format name - testimg1 image01
# 
# Results
#   Returns a string like /Im1, /Im2, etc.
# 
proc getImageName {local} {
    variable  PDF
    set pos [lsearch $PDF(imgList) $local]
    if {$pos < 0} {return ""}
    incr pos
    return [lindex $PDF(imgList) $pos]
}


################################################################
# proc fontObj_Write {subtype base encoding}--
#    Writes a Font object to the disk
# Arguments
#   subtype - The subtype fo this object - /Type1, /Type3, etc
#   base -    The name of the font in the outside world - Arial, Times-Roman, etc
#   encoding - How it's encoded, /MacRomanEncoding for example
#
# Results
#   Updates the xref table,
#   Writes to disk.

proc fontObj_Write {subtype base encoding} {
    variable PDF
    set of $PDF(of)
    # xpdf will accept times, arial, but acroread
    # wants Times-Roman, Arial and Helvetica, etc.
    # What other font names should be here?

    array set fontNames {
        /times      /Times-Roman 
        /helvetica  /Helvetica 
        /arial      		/Arial
        /arial--      		/Arial
        /arial-normal- 		/Arial
        /arial-bold-      	/Arial-Bold
        /arial-bold-italic 	/Arial-BoldItalic
        /arial-italic-	 	/Arial-Italic
	    /times-normal		/Times-Roman
        /helvetica		/Helvetica
        /courier		/Courier
        /symbol			/Symbol
        /times-bold		/Times-Bold
        /helvetica--	/Helvetica
        /helvetica-bold		/Helvetica-Bold
        /courier-bold		/Courier-Bold
        /zapfdingbats		/ZapfDingbats
        /times-italic		/Times-Italic
        /helvetica-italic	/Helvetica-Oblique
        /courier-italic		/Courier-Oblique
        /times-bold-italic	/Times-BoldItalic
        /helvetica-bold-italic	/Helvetica-BoldOblique
        /courier-bold-italic	/Courier-BoldOblique
        arial      		/Arial
        arial--      		/Arial
        arial-normal- 		/Arial
        arial-bold-      	/Arial-Bold
        arial-italic-	 	/Arial-Italic
        arial-bold-italic 	/Arial-BoldItalic
        times-normal		/Times-Roman
        helvetica		/Helvetica
        courier		/Courier
        symbol		/Symbol
        times-bold	/Times-Bold
        helvetica-bold	/Helvetica-Bold
        helvetica--	/Helvetica
        courier-bold	/Courier-Bold
        zapfdingbats	/ZapfDingbats
        times-italic	/Times-Italic
        helvetica-italic	/Helvetica-Oblique
        courier-italic		/Courier-Oblique
        times-bold-italic	/Times-BoldItalic
        helvetica-bold-italic	/Helvetica-BoldOblique
        courier-bold-italic	/Courier-BoldOblique
        }

    if {[info exists fontNames($base)]} {
        set baseFont $fontNames($base)
    } else {
        if {[string first "/" $base] == 0} {
          set baseFont $base
	} else {
          set baseFont /$base
	}
    }
    
    set name /F$PDF(fontNum) 

    set fontObjNum [ObjWrite /Type /Font /Subtype $subtype /Name $name \
           /BaseFont $baseFont  /Encoding $encoding]

    lappend PDF(fontList) $base $name
    set PDF(fontObjNum.$name) $fontObjNum

    incr PDF(fontNum)
    return $name
}


################################################################
# proc ObjWrite {args}--
#    Actually write an object to disk
# Arguments
#   args - a list of key/value pairs to be placed in a dictionary
# 
# Results
#   Updates the xref table,
#   Writes to disk.

proc ObjWrite {args} {
    variable PDF

    set objNum [ObjStart]
    puts $PDF(of) [List2Dict $args]
    ObjEnd
    return $objNum
}

################################################################
# proc Directive_Write {args}--
#    Writes a Directive object to the disk
# Arguments
#   args - key/value pairs
# 
# Results
#   Adds a Directive object (does not contain a dictionary) to disk
#   Updates xref table

proc Directive_Write {args} {
    variable PDF

    set objNum [ObjStart]
    foreach {key val} $args {
        puts $PDF(of) "\[$key $val\]"
    }
    ObjEnd
    return $objNum
}

################################################################
# proc ArrayObj_Write {args} 
#    Writes an array (list) object to the disk
# Arguments
#   args - a list 
# 
# Results
#   Adds a Directive list object (does not contain a dictionary) to disk
#   Updates xref table

proc ArrayObj_Write {args} {
    variable PDF

    set objNum [ObjStart]
    puts $PDF(of) "\[ $args \]"
    ObjEnd
    return $objNum
}

################################################################
# proc initPDF_Write {fileName} --
#    Initialize a PDF file
# Arguments
#   fileName - The name for this pdf document file
# 
# Results
#   A new file is opened.
#   PDF Version identifier is written
#   PDF() values are initialized

proc initPDF_Write {fileName}  {
    variable PDF

    set PDF(of) [open $fileName w]
    fconfigure $PDF(of) -translation lf
    set PDF(idN) 65535
    set PDF(id) f
    set PDF(xrefs) [list [format $pdf::PDF(fmtStr) 0 $pdf::PDF(idN) $pdf::PDF(id)]]
    set PDF(idN) 0
    set PDF(id) n
    set PDF(fontList) ""
    set PDF(imgList) ""
    set PDF(fontNum) 2

    puts $PDF(of) "%PDF-1.4"
}

################################################################
# proc docTree_Write {args}--
#    Writes the document tree definition objects, the Catalog, empty outline,
#    Pages and Page objects
#    Invoked after Font Objects and graphics Stream Object is written.
# Arguments
#   None
# 
# Results
#   PDF document is updated.
# 
proc docTree_Write {args} {
    
    variable PDF

    set pos [llength $PDF(xrefs)]
    set pagecount 1
    set rotate 0
    set pageheight 792
    set pagewidth 612

    foreach {k v} $args {
      set nm [string range $k 1 end]
      set $nm $v
    }
    
    set PDF(root) $pos
    set outline [incr pos]
    set pages [incr pos]
    set kids [incr pos]
    set procset [incr pos $pagecount]
    set contents [incr pos]
    #puts "DocTree: PC: $pagecount -- AR: $args -- KIDS: $kids"

    set fonts ""
    set images ""

    foreach {fam id} $PDF(fontList) {
      lappend fonts $id $PDF(fontObjNum.$id) 0 R
    }

    foreach {name id} $PDF(imgList) {
      lappend images $id $PDF(imageObjNum.$id) 0 R
    }

    if {$rotate} {
       set t $pagewidth; set pagewidth $pageheight; set pageheight $t
    }
    
    set kidt $kids
    for {set i 0} {$i < $pagecount} {incr i} {
      append kidset "                 $kidt 0 R\n"
      incr kidt
    }

    set kidset [string trim $kidset]
    
    set resourceList [list /ProcSet [list $procset 0 R]]
    if {![string equal "" $fonts]} {
      lappend resourceList /Font $fonts
    }

    if {![string equal "" $images]} {
      lappend resourceList /XObject $images
    }
    
    ObjWrite /Type /Catalog /Version /1.4 /Outlines "$outline 0 R" /Pages "$pages 0 R"
    ObjWrite /Type /Outlines /Count 0
    ObjWrite /Type /Pages /Kids "\[$kidset\]" /Count $pagecount
    for {set i 0} {$i < $pagecount} {incr i} {
      ObjWrite /Type /Page /Parent "$pages 0 R" /MediaBox "\[0 0 $pagewidth $pageheight\]" \
        /Contents "$contents 0 R" \
        /Resources $resourceList
	incr contents
    }
}

################################################################
# proc finish_Write {}--
#    Writes the xref table, trailer, and startxref values
#    closes the file
# Arguments
#   NONE
# 
# Results
#   Updates the file and closes it.

proc finish_Write {} {
    variable PDF
    set xrefList $PDF(xrefs)

    set cur [tell $PDF(of)]

    puts $PDF(of) xref
    puts $PDF(of) "0 [llength $xrefList]"
    
    foreach l $xrefList {
        puts $PDF(of) $l
    }
    
    # The trailer dictionary must include 
    #  /Size - the number of entries in the xref list
    #  /Root - the master indirect object - with /Type = /Catalog

    puts $PDF(of) "trailer\n<< /Size [llength $xrefList]"
    puts $PDF(of) "/Root $PDF(root) 0 R\n>>"
    puts $PDF(of) "startxref\n$cur\n%%EOF"
    close $PDF(of)
}

################################################################
# proc pdfRectangle {x y w h outline fill linewidth}--
#   Return a string of commands to define a rectangle.
# Arguments
#   x y - X/Y coords in PDF space
#   w h - Height/Width in PDF space
#   outline/fill - The outside/inside colors, in Tcl values
#   linewidth - width of outline in PDF size.
# Results
#   
# 
proc pdfRectangle {x y w h outline fill linewidth} {

    set outline [Color2PDF $outline]
    append def "$outline RG\n"
    append def "$linewidth w\n"
    if {![string equal "" $fill]} {
      set fill [Color2PDF $fill]
      append def "$fill rg\n"
      # B Will close the path, fill, and then stroke the outline
      set end "B"
    } else {
      # S Will close the path and then stroke the outline, no fill
      set end "S"
    }
    append def "$x $y $w $h re\n"
    append def "$end\n"
    return $def
}

################################################################
# proc pdfPolygon {linewidth outline fill coords} --
#   Return a string of commands to define a rectangle.
# Arguments
#   linewidth - width of outline in PDF size.
#   outline/fill - The outside/inside colors, in Tcl values
#   args - a set of x/y pairs.  Need not be closed set of points.
# Results
#   
proc pdfPolygon {linewidth outline fill coords} {

    if {($fill != "") && ($outline != "")} {
        set outline [Color2PDF $outline]
        append def "$outline RG\n"
        append def "$linewidth w\n"
        
        set fill [Color2PDF $fill]
        append def "$fill rg\n"
        
        set end "B\n"
    } elseif {$fill != ""} {
        set fill [Color2PDF $fill]
        append def "$fill rg\n"
        set end "f\n"
    } elseif {$outline != ""} {
        set outline [Color2PDF $outline]
        append def "$outline RG\n"
        append def "$linewidth w\n"
        set end "S\n"
    }

    set pdfCmd m
    
    foreach {x1 y1} $coords {break;}
    lappend coords $x1 $y1

    foreach {x y} $coords {
      append def "$x $y $pdfCmd\n"
      set pdfCmd l
    }
    append def "$end\n"
    return $def
}

################################################################
# proc pdfLine {linewidth color coords}--
#   Return a string of commands to define a polyline
# Arguments
#   linewidth: Width in PDF coords
#   color: color as Tcl value
#   coords: A list of coords in PDF coords
# Results
#   No side effects

proc pdfLine {linewidth color coords} {
    set color [Color2PDF $color]

    append def "$linewidth w\n"
    append def "$color RG\n"

    set pdfCmd m

    foreach {x y} $coords   {
      append def "$x $y $pdfCmd\n"
      set pdfCmd l
    }    
    append def "S\n"
    return $def
}

################################################################
# proc pdfImageObj {subtype base encoding}--
#    Writes a Font object to the disk
# Arguments
#   subtype - The subtype fo this object - /Type1, /Type3, etc
#   base -    The name of the font in the outside world - Arial, Times-Roman, etc
#   encoding - How it's encoded, /MacRomanEncoding for example
#
# Results
#   Updates the xref table,
#   Writes to disk.

proc pdfImageObj {imgName width height hexStream} {
    variable PDF

    set imageObjNum [streamObj_Write $hexStream /Type /XObject \
        /Subtype /Image /Width $width \
        /Height $height /ColorSpace /DeviceRGB /BitsPerComponent 8 \
	/Filter /ASCIIHexDecode ]
    
    set imageObjID /Im$imageObjNum 
    lappend PDF(imgList) $imgName $imageObjID
    set PDF(imageObjNum.$imageObjID) $imageObjNum

    return $imageObjID
}

# Procedure: pdfArc - Convert the arc item's characteristics
#                         into an appropriate PDF bezier curve.
# Inputs:
#   w    - Path to the canvas widget containing the arc item.
#   item - Identifier of the canvas arc item.
#   scaleMult - a scaling multiplier
#
# Output:
#   Text stream of PDF elements representing the canvas arc item..
#

proc pdfArc {linewidth outline fill xa ya xb yb start extent style} {

    variable var

    set outline [Color2PDF $outline]
    append def "$outline RG\n"
    append def "$linewidth w\n"

    if {![string equal "" $fill]} {
      set fill [Color2PDF $fill]
      append def "$fill rg\n"
      # B Will close the path, fill, and then stroke the outline
      set defEnd "B\n"
    } else {
      # S Will close the path and then stroke the outline, no fill
      set defEnd "S\n"
    }

    append streamdata $def

	# Normalize start to range of -360 degrees to +360 degrees.

	if {$start >= 0} {
	    while {$start >= 360} {
		set start [expr {$start - 360}]
	    }
	} else {
	    while {$start <= -360} {
		set start [expr {$start + 360}]
	    }
	}
	# Normalize extent to range of -360 degrees to +360 degrees.

	if {$extent >= 0} {
	    while {$extent >= 360} {
		set extent [expr {$extent - 360}]
	    }
	    # Break up the arc as necessary.
	    if {$extent > 180} {
		set end [expr {$start + 180}]
		set end2 [expr {$start + $extent}]
	    } else {
		set end [expr {$start + $extent}]
		set end2 {}
	    }
	} else {
	    while {$extent <= -360} {
		set extent [expr {$extent + 360}]
	    }
	    # Break up the arc as necessary.
	    if {$extent < -180} {
		set end [expr {$start - 180}]
		set end2 [expr {$start + $extent}]
	    } else {
		set end [expr {$start + $extent}]
		set end2 {}
	    }
	}

	# Find the center of the fitting ellipse.
	set x [expr {($xa + $xb) / 2}]
	set y [expr {($ya + $yb) / 2}]
	# Find the x/y extens of the fitting ellipse.
	set a [expr {($xb - $xa) / 2}]
	set b [expr {($yb - $ya) / 2}]
	# Find the radian values of the start and end angles.
	set alpha [expr {$start * $var(deg2rad)}]
	set beta [expr {$end * $var(deg2rad)}]
	# Establish the Bezier control point for the arc.
	set bcp [expr {4.0/3.0 * \
			   (1 - cos(($beta - $alpha)/2)) \
			   / sin(($beta - $alpha)/2)}]
	# Calculate sines and cosines of the start and end angles.
	set sin_alpha [expr {sin($alpha)}]
	set cos_alpha [expr {cos($alpha)}]
	set sin_beta [expr {sin($beta)}]
	set cos_beta [expr {cos($beta)}]
	# Find the start point of the Bezier.
	set p0_x [expr {$x + $a * $cos_alpha}]
	set p0_y [expr {$y + $b * $sin_alpha}]
	# Start the new sub-path and current point.
	append streamdata $p0_x " " $p0_y " m\n"
	# Find the control points and end point of the Bezier.
	set p1_x [expr {$x + $a * ($cos_alpha - $bcp * $sin_alpha)}]
	set p1_y [expr {$y + $b * ($sin_alpha + $bcp * $cos_alpha)}]
	set p2_x [expr {$x + $a * ($cos_beta + $bcp * $sin_beta)}]
	set p2_y [expr {$y + $b * ($sin_beta - $bcp * $cos_beta)}]
	set p3_x [expr {$x + $a * $cos_beta}]
	set p3_y [expr {$y + $b * $sin_beta}]
	# Append the cubic Bezier curve and
	# make p3_x/p3_y the new current point.
	append streamdata $p1_x " " $p1_y " " \
	    $p2_x " " $p2_y " " $p3_x " " $p3_y " c\n"
	# Add another cubic Bezier curve if the entire arc > 180 degrees.
	if {$end2 ne {}} {
	    # The original end angle is now the start angle.
	    set alpha $beta
	    # Find the radian value of the new end angle.
	    set beta [expr {$end2 * $var(deg2rad)}]
	    # The original end sine and cosine
	    # are are now the start sine and cosine.
	    set sin_alpha $sin_beta
	    set cos_alpha $cos_beta
	    # Calculate sine and cosine of the new end angle.
	    set sin_beta [expr {sin($beta)}]
	    set cos_beta [expr {cos($beta)}]
	    # Establish the new Bezier control point for the second arc.
	    set bcp [expr {4.0/3.0 * \
			       (1 - cos(($beta - $alpha)/2)) \
			       / sin(($beta - $alpha)/2)}]
	    # Find the control points and end point of the second Bezier.
	    set p4_x [expr {$x + $a * ($cos_alpha - $bcp * $sin_alpha)}]
	    set p4_y [expr {$y + $b * ($sin_alpha + $bcp * $cos_alpha)}]
	    set p5_x [expr {$x + $a * ($cos_beta + $bcp * $sin_beta)}]
	    set p5_y [expr {$y + $b * ($sin_beta - $bcp * $cos_beta)}]
	    set p6_x [expr {$x + $a * $cos_beta}]
	    set p6_y [expr {$y + $b * $sin_beta}]
	    # Append the second cubic Bezier curve and
	    # make p6_x/p6_y the new current point.
	    append streamdata $p4_x " " $p4_y " " \
		$p5_x " " $p5_y " " $p6_x " " $p6_y " c\n"
	}

	# Determine completion of shape for the appropriate arc style.
	switch -- $style  {
	    "" -
	    pieslice {
            # Add lines for wedge of pieslice
            append streamdata $x " " $y " l % Line to center of slice\n"
            append streamdata "h % Close the pieslice\n"
            # append streamdata [lindex $rendering 1] "\n"
            append streamdata $defEnd
	    }
	    arc {
            # An arc will only be stroked.
            append streamdata "S\n"
	    }
	    chord {
            # Draw a straight line connecting the ends of the arc.
            append streamdata "h % Close the chord\n"
            # append streamdata [lindex $rendering 1] "\n"
            append streamdata $defEnd
	    }
	}

    return $streamdata
}

# Procedure: pdfOval - Convert the oval item's characteristics
#                          into a PDF bezier curve element.
# Inputs:
#   w    - Path to the canvas widget containing the oval item.
#   item - Identifier of the canvas oval item.
#
# Output:
#   Text stream of PDF elements representing the canvas oval item.
#
# proc pdfOval {w item scaleMult} {}

proc pdfOval {xa ya xb yb outline fill linewidth} {
	variable var

    set outline [Color2PDF $outline]
    append streamdata "\n$outline RG\n"
    append streamdata "\n$linewidth w\n"

    if {![string equal "" $fill]} {
        set fill [Color2PDF $fill]
        append streamdata "\n$fill rg\n"
        # B Will close the path, fill, and then stroke the outline
        set end "\nB\n"
    } else {
        # S Will close the path and then stroke the outline, no fill
        set end "S\n"
    }

    # Find center point of the oval.
	set xc [expr {(($xa + $xb) / 2.0)}]
	set yc [expr {(($ya + $yb) / 2.0)}]
	# Find the major and minor axes.
	set a [expr {(($xb - $xa) / 2.0) }]
	set b [expr {(($yb - $ya) / 2.0) }]
	# Make control points offsets
	set ak [expr {$a * $var(kappa)}]
	set bk [expr {$b * $var(kappa)}]
	# Set new current point 
	append streamdata $xa " " $yc " m\n"
	append streamdata $xa " " [expr {$yc + $bk}] " " \
	    [expr {$xc - $ak}] " " $yb " " $xc " " $yb " c\n"
	append streamdata [expr {$xc + $ak}] " " $yb " " \
	    $xb " " [expr {$yc + $bk}] " " $xb " " $yc " c\n"
	append streamdata $xb " " [expr {$yc - $bk}] " " \
	    [expr {$xc + $ak}] " " $ya " " $xc " " $ya " c\n"
	append streamdata [expr {$xc - $ak}] " " $ya " " \
	    $xa " " [expr {$yc - $bk}] " " $xa " " $yc " c\n"
    
    append streamdata $end

    return $streamdata
}

proc pdfRawCurve {linewidth outline fill coords} {
    
    if {($fill != "") && ($outline != "")} {
        set outline [Color2PDF $outline]
        append def "$outline RG\n"
        append def "$linewidth w\n"
        
        set fill [Color2PDF $fill]
        append def "$fill rg\n"
        
        set end "B\n"
    } elseif {$fill != ""} {
        set fill [Color2PDF $fill]
        append def "$fill rg\n"
        set end "f\n"
    } elseif {$outline != ""} {
        set outline [Color2PDF $outline]
        append def "$outline RG\n"
        append def "$linewidth w\n"
        set end "S\n"
    }
    
    if {$fill != ""} {    
        if {([lindex $coords 0] != [lindex $coords end-1]) \
                || ([lindex $coords 1] != [lindex $coords end])} {
            lappend coords [lindex $coords 0] [lindex $coords 1]
        }
    }
   
    set length [llength $coords]
    set borrow [expr {(($length / 2) - 5) % 3}]
    
    if {$borrow == 0} {
        lappend coords [lindex $coords 2] [lindex $coords 3]
    }

    append def "[lindex $coords 0] [lindex $coords 1] m\n"

    for {set i 0} {$i <= $length} {incr i 6} {
        foreach {x y x1 y1 x2 y2 x3 y3} [lrange $coords $i [expr {$i + 7}]] {
            if {($x == $x1) && ($y == $y1) && ($x2 == $x3) && ($y2 == $y3)} {
                append def "$x3 $y3 l\n"
            } else {
                append def "$x $y $x1 $y1 $x2 $y2 $x3 $y3 c\n"
            }
        }
    }
    
    append def "$end\n"
    
    return $def
}

# tkTrig:1020
proc pdfBezier {linewidth outline fill coords} {
    
    if {($fill != "") && ($outline != "")} {
        set outline [Color2PDF $outline]
        append def "$outline RG\n"
        append def "$linewidth w\n"
        
        set fill [Color2PDF $fill]
        append def "$fill rg\n"
        
        set end "B\n"
    } elseif {$fill != ""} {
        set fill [Color2PDF $fill]
        append def "$fill rg\n"
        set end "f\n"
    } elseif {$outline != ""} {
        set outline [Color2PDF $outline]
        append def "$outline RG\n"
        append def "$linewidth w\n"
        set end "S\n"
    }
    
    if {$fill != ""} {    
        if {([lindex $coords 0] != [lindex $coords end-1]) \
                || ([lindex $coords 1] != [lindex $coords end])} {
            lappend coords [lindex $coords 0] [lindex $coords 1]
        }
    }

    if {[lindex $coords 0] == [lindex $coords end-1] && \
                [lindex $coords 1] == [lindex $coords end]} {
        set closed 1

        set controlx [expr {0.5  * [lindex $coords end-3] \
                + 0.5  *[lindex $coords 0]}]
        set controly [expr {0.5  * [lindex $coords end-2] \
                + 0.5  *[lindex $coords 1]}]
        set controlx1 [expr {0.167* [lindex $coords end-3] \
                + 0.833*[lindex $coords 0]}]
        set controly1 [expr {0.167* [lindex $coords end-2] \
                + 0.833*[lindex $coords 1]}]
        set controlx2 [expr {0.833* [lindex $coords 0] \
                + 0.167*[lindex $coords 2]}]
        set controly2 [expr {0.833* [lindex $coords 1] \
                + 0.167*[lindex $coords 3]}]
        set controlx3 [expr {0.5  * [lindex $coords 0] \
                + 0.5  *[lindex $coords 2]}]
        set controly3 [expr {0.5  * [lindex $coords 1] \
                + 0.5  *[lindex $coords 3]}]
        
        append def "$controlx $controly m\n $controlx1 $controly1 $controlx2 \
                $controly2 $controlx3 $controly3 c\n"
    } else {
        set closed 0
        
        set controlx3 [lindex $coords 0]
        set controly3 [lindex $coords 1]
        
        append def "$controlx3 $controly3 m\n"
    }
    
    set length [llength $coords]

    for {set i 2} {$i < ($length - 2)} {incr i 2} {
        foreach {pointx pointy pointx1 pointy1} \
                [lrange $coords $i [expr {$i + 3}]] break

        set controlx1 [expr {0.333 * $controlx3 + 0.667 * $pointx}]
        set controly1 [expr {0.333 * $controly3 + 0.667 * $pointy}]

        if {($i == ($length - 4)) && (!$closed)} {
            set controlx3 $pointx1
            set controly3 $pointy1
        } else {
            set controlx3 [expr {0.5 * $pointx + 0.5 * $pointx1}]
            set controly3 [expr {0.5 * $pointy + 0.5 * $pointy1}]
        }

        set controlx2 [expr {0.333 * $controlx3 + 0.667 * $pointx}]
        set controly2 [expr {0.333 * $controly3 + 0.667 * $pointy}]

        append def "$controlx1 $controly1 $controlx2\
                $controly2 $controlx3 $controly3 c\n"
    }
    
    append def $end
    
    return $def
}


}

#pdf::initPDF_Write moj.pdf 

#set n1 [pdf::fontObj_Write /Type1 Times-Roman /MacRomanEncoding] 
#append geom [pdf::pdfText $n1 21 red 20 540 "Red'Large Helvetica"]

#pdf::docTree_Write 
#pdf::Directive_Write /PDF /Text 
#pdf::pageObj_Write $geom
#pdf::finish_Write

if 0 {
if {[lsearch $argv "-test"] >= 0} {
puts "COLORS: Gray50: [pdf::Color2PDF gray50] black: [pdf::Color2PDF black] #888: [pdf::Color2PDF #888]"
puts "Dict1: \n[pdf::List2Dict {/One 1 /Two 2}]"
puts "Nested Dict: \n[pdf::List2Dict {
  /L1K1 L1One
    /Ind1 {
      /L2K1 L2One 
      /L2K2 {a b c}
      /L2K3 L2Two
        /Ind2 {
          /L3K1 L3One
          /L3K2 L3Two
        } 
      } 
   }] "

pdf::initPDF_Write twobox.pdf 
append geom [pdf::pdfRectangle 429 150 80 100 red green 4]
append geom [pdf::pdfRectangle 10 10 30 50 blue yellow 8] 
pdf::fontObj_Write /Type1 Helvetica /MacRomanEncoding 
pdf::docTree_Write 
pdf::Directive_Write /PDF /Text 
pdf::pageObj_Write $geom
pdf::finish_Write


pdf::initPDF_Write textonly.pdf 
set n1 [pdf::fontObj_Write /Type1 Helvetica /MacRomanEncoding] 
set n2 [pdf::fontObj_Write /Type1 Arial /MacRomanEncoding] 
set n3 [pdf::fontObj_Write /Type1 Times-Roman /MacRomanEncoding] 
append txt [pdf::pdfText $n1 24 red 20 40 "Red Large Helvetica"]
append txt [pdf::pdfText $n2 18 green 20 60 "Green Medium Arial"]
append txt [pdf::pdfText $n3 12 #00a0ff 20 80 "Blue Small Times"]
pdf::docTree_Write 
pdf::Directive_Write /PDF /Text 
pdf::pageObj_Write $txt
pdf::finish_Write


set geom "" 
set txt ""
pdf::initPDF_Write twoboxNText.pdf 
append geom [pdf::pdfRectangle 429 150 80 100 red green 4]
append geom [pdf::pdfRectangle 10 10 30 50 blue yellow 8] 
append geom [pdf::pdfPolygon 4 pink orange {70 10 90 10 80 50} ] 
set n1 [pdf::fontObj_Write /Type1 Helvetica /MacRomanEncoding] 
set n2 [pdf::fontObj_Write /Type1 Arial /MacRomanEncoding] 
set n3 [pdf::fontObj_Write /Type1 Times-Roman /MacRomanEncoding] 
append txt [pdf::pdfText $n1 24 red 20 140 "Red Large Helvetica"]
append txt [pdf::pdfText $n2 18 green 20 160 "Green Medium Arial"]
append txt [pdf::pdfText $n3 12 #00a0ff 20 180 "Blue Small Times"]
pdf::docTree_Write 
pdf::Directive_Write /PDF /Text 

append geom \n $txt
pdf::pageObj_Write $geom
pdf::finish_Write
}


}
